package com.codenic.codenic_alarm

import android.content.ComponentName
import android.content.Context
import android.content.Intent
import android.content.ServiceConnection
import android.os.IBinder
import androidx.core.app.NotificationManagerCompat

import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.plugin.common.BasicMessageChannel
import io.flutter.plugin.common.JSONMessageCodec
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result

/**
 * A Flutter plugin that starts an alarm service to play the alarm ringtone and show a sticky
 * notification when the alarm is triggered.
 */
class CodenicAlarmPlugin: FlutterPlugin, MethodCallHandler {
  class Methods {
    companion object {
      const val START_ALARM = "startAlarm"
      const val DISMISS_ALARM = "dismissAlarm"
      const val GET_ACTIVE_ALARM = "getActiveAlarm"
    }
  }

  /**
   * The [MethodChannel] that is used to invoked methods in the plugin from the Flutter side
   * which include starting and dismissing the alarm.
   */
  private lateinit var channel : MethodChannel

  /**
   * The [BasicMessageChannel] that is used to send events from the Android side to the Flutter
   * side, typically to inform that the alarm has been started or dismissed.
   */
  private lateinit var messageChannel : BasicMessageChannel<Any>

  private lateinit var context : Context

  /**
   * Indicates whether the plugin is currently bound to the [AlarmService].
   *
   * This prevents any exception that may occur when the plugin tries to unbind from the service
   * when it is not bound.
   */
  private var isAlarmServiceBinded = false

  /**
   * A [ServiceConnection] that is used to listen to events from the [AlarmService] then forwards
   * them to the Flutter side via the [messageChannel].
   */
  private val serviceConnection = object : ServiceConnection {
    override fun onServiceConnected(name: ComponentName, service: IBinder) {
      isAlarmServiceBinded = true

      val binder = service as AlarmService.LocalBinder
      val alarmService = binder.getService()

      alarmService.setAlarmServiceListener(
        object : AlarmServiceListener {
          override fun onAlarmStarted(alarmInfo: AlarmInfo) {
            messageChannel.send(
              mapOf(
                "event" to "alarm_started",
                "alarm" to alarmInfo.toJson(),
              )
            )
          }
          override fun onAlarmDismissed() {
            messageChannel.send(mapOf("event" to "alarm_dismissed"))
          }
        }
      )
    }

    override fun onServiceDisconnected(name: ComponentName) {
      isAlarmServiceBinded = false
    }
  }

  override fun onAttachedToEngine(flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
    channel = MethodChannel(flutterPluginBinding.binaryMessenger, "codenic_alarm")
    channel.setMethodCallHandler(this)

    messageChannel = BasicMessageChannel(
      flutterPluginBinding.binaryMessenger,
      "codenic_alarm/message_channel",
      JSONMessageCodec.INSTANCE,
    )

    context = flutterPluginBinding.applicationContext
  }

  override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
    channel.setMethodCallHandler(null)

    if (isAlarmServiceBinded) {
      context.unbindService(serviceConnection)
    }
  }

  override fun onMethodCall(call: MethodCall, result: Result) {
    when (call.method) {
      Methods.START_ALARM -> {
        return onStartAlarm(call, result)
      }
      Methods.DISMISS_ALARM -> {
        return onDismissAlarm(result)
      }
      Methods.GET_ACTIVE_ALARM -> {
        return onGetActiveAlarm(result)
      }
      else -> result.notImplemented()
    }
  }

  /**
   * Starts the alarm service to play the alarm ringtone and show a sticky notification.
   *
   * If the user has disabled notifications, this method will return an error.
   */
  private fun onStartAlarm(call: MethodCall, result: Result) {
    val areNotificationsEnabled =
      NotificationManagerCompat.from(context).areNotificationsEnabled()

    if (!areNotificationsEnabled) {
      return result.error(
        "notifications-disabled",
        "Cannot play alarm because notifications are disabled.",
        null,
      )
    }

    // Bind the plugin to the `AlarmService` to listen for events from the service once its started

    val bindServiceIntent = Intent(context, AlarmService::class.java)

    val bindResult = context.bindService(bindServiceIntent, serviceConnection, 0)

    if (!bindResult) {
      return result.error(
        "bind-service-failed",
        "Failed to bind to the alarm service.",
        null,
      )
    }

    // Start the alarm service

    val startServiceIntent = Intent(context, AlarmService::class.java)

    startServiceIntent.action = AlarmService.StartAlarmIntent.ACTION

    startServiceIntent.putExtra(
       AlarmService.StartAlarmIntent.EXTRA_ALARM_ID,
      call.argument<Int>(AlarmService.StartAlarmIntent.EXTRA_ALARM_ID),
    )

    startServiceIntent.putExtra(
      AlarmService.StartAlarmIntent.EXTRA_ALARM_TITLE,
      call.argument<String>(AlarmService.StartAlarmIntent.EXTRA_ALARM_TITLE),
    )

    startServiceIntent.putExtra(
      AlarmService.StartAlarmIntent.EXTRA_ALARM_CONTENT,
      call.argument<String>(AlarmService.StartAlarmIntent.EXTRA_ALARM_CONTENT),
    )

    context.startService(startServiceIntent)
      ?: return result.error(
        "start-service-failed",
        "Failed to start the alarm service.",
        null,
      )

    return result.success(null)
  }

  /**
   * Dismisses the alarm service, effectively stopping the alarm ringtone and hiding the sticky
   * notification.
   */
  private fun onDismissAlarm(result: Result) {
    val serviceIntent = Intent(context, AlarmService::class.java)
    serviceIntent.action = AlarmService.DismissAlarmIntent.ACTION

    context.startService(serviceIntent)

    return result.success(null)
  }

  /**
   * Returns the currently active alarm, if any.
   */
  private fun onGetActiveAlarm(result: Result) {
    return result.success(AlarmService.getInstance()?.alarmInfo?.toJson())
  }
}
