import 'dart:async';
import 'dart:isolate';
import 'dart:ui';

import 'package:android_alarm_manager_plus/android_alarm_manager_plus.dart';
import 'package:codenic_alarm/codenic_alarm.dart';
import 'package:codenic_alarm/src/codenic_alarm_interface.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

/// {@template CodenicAlarm}
/// A Flutter plugin to schedule alarms on Android and iOS.
/// {@endtemplate}
class CodenicAlarm implements CodenicAlarmInterface {
  /// {@macro CodenicAlarm}
  ///
  /// This returns a singleton instance of [CodenicAlarm].
  factory CodenicAlarm() => _instance;

  CodenicAlarm._internal();

  static final CodenicAlarm _instance = CodenicAlarm._internal();

  /// The name of the [SendPort] used by this plugin to send the currently
  /// active [AlarmInfo] to its corresponding [ReceivePort] from the
  /// [CodenicAlarmOverlayHandler].
  static const alarmInfoSendPortName = 'codenic_alarm/alarm_info_send_port';

  /// The [MethodChannel] used to invoke commands to the platform side of this
  /// plugin.
  @visibleForTesting
  MethodChannel get methodChannel => const MethodChannel('codenic_alarm');

  /// The [BasicMessageChannel] used to receive events from the platform side
  /// of this plugin.
  BasicMessageChannel<Object?> get messageChannel => const BasicMessageChannel(
        'codenic_alarm/message_channel',
        JSONMessageCodec(),
      );

  /// Indicates whether the [initialize] methods has been called.
  bool _isInitialized = false;

  @override
  FutureOr<bool> initialize() async {
    if (_isInitialized) {
      return true;
    }

    _isInitialized = true;

    await AndroidAlarmManager.initialize();
    _initializeMessageChannelHandler();

    return true;
  }

  /// Initializes the [messageChannel] handler by forwarding the received
  /// messages from the platform channel to the [CodenicAlarmOverlayHandler].
  void _initializeMessageChannelHandler() =>
      messageChannel.setMessageHandler((message) async {
        switch (message) {
          case {'event': 'alarm_started'}:
            IsolateNameServer.lookupPortByName(
              CodenicAlarm.alarmInfoSendPortName,
            )?.send(message['alarm']);

          case {'event': 'alarm_dismissed'}:
            IsolateNameServer.lookupPortByName(
              CodenicAlarm.alarmInfoSendPortName,
            )?.send(null);
        }

        return message;
      });

  @override
  FutureOr<bool> schedule({
    required Duration delay,
    required AlarmInfo alarmInfo,
  }) =>
      scheduleAt(schedule: DateTime.now().add(delay), alarmInfo: alarmInfo);

  @override
  FutureOr<bool> scheduleAt({
    required DateTime schedule,
    required AlarmInfo alarmInfo,
  }) =>
      AndroidAlarmManager.oneShotAt(
        schedule,
        alarmInfo.id,
        _alarmCallback,
        allowWhileIdle: true,
        alarmClock: true,
        exact: true,
        rescheduleOnReboot: true,
        wakeup: true,
        params: alarmInfo.toJson(),
      );

  @override
  FutureOr<void> dismissAlarm() =>
      CodenicAlarm().methodChannel.invokeMethod('dismissAlarm');

  @override
  FutureOr<bool> cancel({required int id}) => AndroidAlarmManager.cancel(id);

  @override
  FutureOr<AlarmInfo?> activeAlarm() async {
    final alarmData =
        await methodChannel.invokeMapMethod<String, dynamic>('getActiveAlarm');

    if (alarmData == null) {
      return null;
    }

    return AlarmInfo.fromJson(alarmData);
  }
}

/// The callback function invoked to play an alarm.
@pragma('vm:entry-point')
Future<void> _alarmCallback(int id, Map<String, dynamic> data) async {
  final codenicAlarm = CodenicAlarm();

  await codenicAlarm.initialize();
  await codenicAlarm.methodChannel.invokeMethod('startAlarm', data);

  IsolateNameServer.lookupPortByName(CodenicAlarm.alarmInfoSendPortName)
      ?.send(data);
}
